@@ -63,13 +63,13 @@ static int makeit(dos_name_t *dosname,
 {
 	Stream_t *Target;
 	CreateArg_t *arg = (CreateArg_t *) arg0;
-	int fat;
-	direntry_t subEntry;	
+	uint32_t fat;
+	direntry_t subEntry;
 
 	/* will it fit? At least one cluster must be free */
 	if (!getfreeMinClusters(targetEntry->Dir, 1))
 		return -1;
-	
+
 	mk_entry(dosname, ATTR_DIR, 1, 0, arg->mtime, &targetEntry->dir);
 	Target = OpenFileByDirentry(targetEntry);
 	if(!Target){
@@ -89,7 +89,7 @@ static int makeit(dos_name_t *dosname,
 	mk_entry_from_base("..      ", ATTR_DIR, fat, 0, arg->mtime, &subEntry.dir);
 	dir_write(&subEntry);
 
-	FLUSH((Stream_t *) Target);
+	FLUSH(Target);
 	subEntry.entry = 0;
 	GET_DATA(Target, 0, 0, 0, &fat);
 	mk_entry_from_base(".       ", ATTR_DIR, fat, 0, arg->mtime, &subEntry.dir);
@@ -148,7 +148,7 @@ static int createDirCallback(direntry_t *entry UNUSEDP, MainParam_t *mp)
 		FREE(&ret);
 		return GOT_ONE;
 	}
-	
+
 }
 
 void mmd(int argc, char **argv, int type UNUSEDP) NORETURN;
@@ -172,7 +172,7 @@ void mmd(int argc, char **argv, int type UNUSEDP)
 			case '?':
 				usage(1);
 			case 'o':
-				handle_clash_options(&arg.ch, c);
+				handle_clash_options(&arg.ch, (char) c);
 				break;
 			case 'D':
 				if(handle_clash_options(&arg.ch, *optarg))
